unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, shlobj, ActiveX, ExtCtrls;

type
  TColLimits = record
    loR,
      loG,
      loB,
      hiR,
      hiG,
      hiB: Byte;
  end;

type
  TForm1 = class(TForm)
    SpeedButtonStartClr: TSpeedButton;
    SpeedButtonEndClr: TSpeedButton;
    EditDir: TEdit;
    LabelStartClr: TLabel;
    LabelEndClr: TLabel;
    ButtonBrowse: TButton;
    Label3: TLabel;
    ColorDialog1: TColorDialog;
    ButtonStep1: TButton;
    ButtonStep2: TButton;
    Panel1: TPanel;
    ListBoxStatus: TListBox;
    ListBoxFiles: TListBox;
    LabelMin: TLabel;
    LabelMax: TLabel;
    LabelMinVal: TLabel;
    LabelMaxVal: TLabel;
    ButtonStop: TButton;
    procedure SpeedButtonStartClrClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SpeedButtonEndClrClick(Sender: TObject);
    procedure ButtonBrowseClick(Sender: TObject);
    procedure ButtonStep1Click(Sender: TObject);
    procedure DebugPrint(s: string);
    procedure ButtonStep2Click(Sender: TObject);
    procedure ButtonStopClick(Sender: TObject);
  public
    { Public declarations }
  end;

type
  TWorkThread = class(TThread)
  private
    Items: TStrings;
    Parent: TForm1;
    gmin,
      gmax: Integer;
    globLimits: TColLimits;
    FKillme: Boolean;
    ShortFname,
      LongFname: string;
  protected
    procedure Execute; override;
    procedure SyncedExecute;
  public
    procedure Suicide;
    constructor Create(itms: TStrings; prnt: TForm1; min, max: Integer);
  end;

var
  Form1: TForm1;
  WorkThread1: TWorkThread;

implementation

{$INCLUDE PngSupport.pas}

{$R *.DFM}

type
  PShort = ^SmallInt;
  TColArr = array[0..0] of COLORREF;
  PColArr = ^TColArr;

const
  PngLibname = 'libpng.dll';
  maxval: Integer = $FFFFFFF;

function GetCols(min, max: Integer; limits: TColLimits; var num: Integer): PColArr;
(******************************************************************************
  Create and fill a color table depending on the minimum and maximum value in
  the IMG file.
 ******************************************************************************)
var
  stepR, stepB, stepG: Real;
  currR, currG, currB: Real;
  i: Integer;
begin
  num := max - min;
// Allocate buffer
  GetMem(result, num * sizeof(COLORREF));
  if Assigned(result) then
  begin
// Calc limits
    stepR := (Integer(limits.hiR) - Integer(limits.loR)) / num;
    stepG := (Integer(limits.hiG) - Integer(limits.loG)) / num;
    stepB := (Integer(limits.hiB) - Integer(limits.loB)) / num;
    currR := limits.loR;
    currG := limits.loG;
    currB := limits.loB;
// Fill array
    for i := 0 to num - 1 do
    begin
      currR := currR + stepR;
      currG := currG + stepG;
      currB := currB + stepB;
      result^[i] := RGB(round(currR), round(currG), round(currB));
    end;
  end;
end;

function ProcessIMG_File(fname: string; min, max: Integer; const Limits: TColLimits): Boolean;
(******************************************************************************
  Processes an IMG file and outputs it as PNG!
 ******************************************************************************)
var
  hFile, hMap: THandle;
  fsize: DWORD;
  mFile, SavedmFile: PShort;
  valu, numofelems, x, y, i, lmin, lmax: Integer;
  bmp1: TBitmap;
  cols: PColArr;
  f: Text;
begin
  lmin := maxval;
  lmax := -lmin;
  bmp1 := TBitmap.Create;
  bmp1.Width := 1380;
  bmp1.Height := 1450;
// Open, map, view, unview, unmap, close file ...
  hFile := CreateFile(@fname[1], GENERIC_READ, 0, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
  if hFile <> INVALID_HANDLE_VALUE then
  try
    fsize := GetFileSize(hFile, nil);
    hMap := CreateFileMapping(hFile, nil, PAGE_READONLY, 0, fsize, nil);
    if hMap <> 0 then
    try
      mFile := MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, fsize);
      SavedmFile := mFile;
      if mFile <> nil then
      try
// Find the min/max inside the IMG
        for y := 0 to bmp1.Height - 1 do
          for x := 0 to bmp1.Width - 1 do
          begin
            if mFile^ <> -9999 then
            begin
              if mFile^ <= lmin then lmin := mFile^;
              if mFile^ >= lmax then lmax := mFile^;
            end;
            inc(mFile);
          end;

// If the min/max values are not given, take file-local ones ...
        if (min = maxval) and (max = -min) then
        begin
          min := lmin;
          max := lmax;
        end;

// Create an according color table
        cols := GetCols(min, max, Limits, numofelems);
        if Assigned(cols) then
        try

          mFile := SavedmFile;
// Fill the bitmap
          for y := 0 to bmp1.Height - 1 do
          begin
            for x := 0 to bmp1.Width - 1 do
            begin
              if mFile^ = -9999 then
              begin
                bmp1.Canvas.Pixels[x, y] := clWhite
              end
              else
              begin
                valu := Integer(mFile^) - min;
                if valu < 0 then valu := 0;
                if valu >= numofelems then valu := numofelems - 1;
                bmp1.Canvas.Pixels[x, y] := cols^[valu];
              end;
              inc(mFile);
            end;
          end;
// Only output minimum and maximum values and color table
          AssignFile(f, fname + '.colors');
          Rewrite(f);
          for x := 0 to numofelems - 1 do
            Writeln(f, Format('%d = #%2.2X%2.2X%2.2X', [x + min, GetRValue(cols^[x]), GetGValue(cols^[x]), GetBValue(cols^[x])]));
          CloseFile(f);

// Add some kind of legend ;)
          bmp1.Canvas.Rectangle(5, bmp1.Height - 20 - numofelems div 2 - 5, 120, bmp1.Height - 1);
          x := bmp1.Height - 20 - (numofelems div 2);
          inc(x, 12);
          bmp1.Canvas.TextOut(60, x, 'Local min:');
          inc(x, 12);
          bmp1.Canvas.TextOut(60, x, Format('%d', [lmin]));
          inc(x, 12);
          inc(x, 12);
          bmp1.Canvas.TextOut(60, x, 'Local max:');
          inc(x, 12);
          bmp1.Canvas.TextOut(60, x, Format('%d', [lmax]));

          bmp1.Canvas.Pen.Color := clBlack;
          bmp1.Canvas.MoveTo(58, bmp1.Height - 20 - (numofelems div 2) - 5);
          bmp1.Canvas.LineTo(58, bmp1.Height - 1);
          for i := 1 to numofelems do
            if Odd(i) then
            begin
              y := bmp1.Height - 20 - (i div 2);
              bmp1.Canvas.Pen.Color := cols^[i];
              bmp1.Canvas.MoveTo(10, y);
              bmp1.Canvas.LineTo(30, y);

              if (i = 1) or (i = numofelems) or (i = numofelems div 2) or (i = numofelems div 2 + 1) then
                bmp1.Canvas.TextOut(32, y - 2, Format('%d', [i + min]));

            end;

        finally
          FreeMem(cols);
        end;
      finally
        UnmapViewOfFile(SavedmFile);
      end;
    finally
      CloseHandle(hMap);
    end;
  finally
    CloseHandle(hFile);
  end;
  result := SavehBmp2PngFile(fname + '.png', bmp1.Handle, bmp1.Width, bmp1.Height, COLORREF(-1), 'http://assarbad.net');
// Free memory
  bmp1.FreeImage;
  bmp1.Destroy;
end;

function BrowseForFolderW(hwnd: HWND; szTitle: PWideChar): WideString;
var
  biw: BROWSEINFOW;
  folder: array[0..MAX_PATH] of WideChar;
  pidl: PItemIDList;
  ppMalloc: IMalloc;
begin
  result := '';
  ZeroMemory(@biw, sizeof(biw));
  if SHGetMalloc(ppMalloc) = NOERROR then
  try
    with biw do
    begin
      hwndOwner := hwnd;
      pszDisplayName := @folder[0];
      lpszTitle := PWideChar(szTitle);
      ulFlags := BIF_RETURNONLYFSDIRS;
    end;
    pidl := SHBrowseForFolderW(biw);
    if Assigned(pidl) then
    try
      if SHGetPathFromIDListW(pidl, folder) then
        SetString(result, folder, lstrlenW(folder));
    finally
      ppMalloc.Free(pidl);
    end;
  finally
    ppMalloc._Release;
  end;
end;

function ReadMinMax(fname: string; var min: Integer; var max: Integer): Boolean;
var
  t: Text;
  s: string;
  foundmin: Boolean;
  foundmax: Boolean;
const
  maxc = 'max. value  : ';
  minc = 'min. value  : ';
begin
  foundmin := False;
  foundmax := False;
  AssignFile(t, fname);
  Reset(t);
  while not EOF(t) do
  begin
    Readln(t, s);
    if pos(minc, s) > 0 then
    begin
      s := copy(s, Length(minc) + 1, length(s));
      min := StrToInt(s);
      foundmin := True;
    end;
    if pos(maxc, s) > 0 then
    begin
      s := copy(s, Length(maxc) + 1, length(s));
      max := StrToInt(s);
      foundmax := True;
    end;
  end;
  CloseFile(t);
  result := foundmin and foundmax;
end;

constructor TWorkThread.Create(itms: TStrings; prnt: TForm1; min, max: Integer);
begin
  Items := itms;
  Parent := prnt;
  gmin := min;
  gmax := max;
  FKillme := False;
  inherited Create(False);
end;

procedure TWorkThread.Suicide;
begin
  Parent.DebugPrint('WARNING: Now committing suicide ...');
  FKillme := True;
end;

procedure TWorkThread.SyncedExecute;
begin
  Application.ProcessMessages;
  if ProcessIMG_File(LongFName, gmin, gmax, globLimits) then
    Parent.DebugPrint(Format('Saved IMG as PNG: %s', [ChangeFileExt(ShortFname, '.png')]));
  Application.ProcessMessages;
end;

procedure TWorkThread.Execute;
var
  i: Integer;
begin
  FKillme := False;
  with globLimits do
  begin
    globLimits.loR := GetRValue(Parent.LabelStartClr.Color);
    globLimits.loG := GetGValue(Parent.LabelStartClr.Color);
    globLimits.loB := GetBValue(Parent.LabelStartClr.Color);
    Parent.DebugPrint(Format('Color.Lo: %2.2X%2.2X%2.2X', [globLimits.loR, globLimits.loG, globLimits.loB]));

    globLimits.hiR := GetRValue(Parent.LabelEndClr.Color);
    globLimits.hiG := GetGValue(Parent.LabelEndClr.Color);
    globLimits.hiB := GetBValue(Parent.LabelEndClr.Color);
    Parent.DebugPrint(Format('Color.Hi: %2.2X%2.2X%2.2X', [globLimits.hiR, globLimits.hiG, globLimits.hiB]));
  end;

  for i := 0 to Items.Count - 1 do
  begin
    ShortFname := Items[i];
    LongFname := Parent.EditDir.Text + '\' + ShortFname;

    Parent.DebugPrint('------------------------------------------');
    Parent.DebugPrint(Format('Processing file %s (min: %d - max: %d)', [ShortFname, gmin, gmax]));


    if FileExists(LongFname) then
      Synchronize(SyncedExecute);

    if FKillme then
      Exit;
  end;

  if Terminated then Exit;
end;

procedure TForm1.DebugPrint(s: string);
begin
  ListBoxStatus.Items.Append(s);
  ListBoxStatus.TopIndex := ListBoxStatus.Items.Count - 1;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  LabelStartClr.Color := clRed;
  LabelEndClr.Color := clGreen;
  EditDir.Text := GetCurrentDir;
  ButtonStep2.Enabled := False;
  DebugPrint('Initialized successfully ...');
end;

procedure TForm1.SpeedButtonStartClrClick(Sender: TObject);
begin
  if ColorDialog1.Execute then
  begin
    LabelStartClr.Color := ColorDialog1.Color;
  end;
end;

procedure TForm1.SpeedButtonEndClrClick(Sender: TObject);
begin
  if ColorDialog1.Execute then
  begin
    LabelEndClr.Color := ColorDialog1.Color;
  end;
end;

procedure TForm1.ButtonBrowseClick(Sender: TObject);
var
  s: string;
begin
  (Sender as TButton).Enabled := False;
  s := BrowseForFolderW(self.Handle, 'Choose a working folder');
  if s <> '' then
  begin
    ListBoxFiles.Clear;
    EditDir.Text := s;
    ButtonStep2.Enabled := False;
    ButtonStep1.Enabled := True;
  end;
  (Sender as TButton).Enabled := True;
end;

procedure TForm1.ButtonStep1Click(Sender: TObject);
var
  SearchRec: TSearchRec;
  i, min, max,
    gmin, gmax: Integer;
begin
  ListBoxStatus.Clear;
  (Sender as TButton).Enabled := False;
  gmin := 9999;
  gmax := -9999;
  ListBoxFiles.Clear;
  if FindFirst(EditDir.Text + '\*.doc', faAnyFile, SearchRec) = 0 then
  begin
    repeat
      ListBoxFiles.Items.Append(SearchRec.Name);
    until FindNext(SearchRec) <> 0;
  end
  else
  begin
    Application.MessageBox('Could not find any .DOC file', nil, 0);
    Exit;
  end;
  for i := 0 to ListBoxFiles.Items.Count - 1 do
  begin
    if ReadMinMax(EditDir.Text + '\' + ListBoxFiles.Items[i], min, max) then
    begin
      if min < gmin then gmin := min;
      if max > gmax then gmax := max;
    end;
  end;
  DebugPrint(Format('Min: %.4d - Max: %.4d', [gmin, gmax]));
  for i := 0 to ListBoxFiles.Items.Count - 1 do
    ListBoxFiles.Items[i] := ChangeFileExt(ListBoxFiles.Items[i], '.img');
  LabelMinVal.Caption := IntToStr(gmin);
  LabelMaxVal.Caption := IntToStr(gmax);
  DebugPrint('You may proceed with step 2!');
  ButtonStep2.Enabled := True;
  (Sender as TButton).Enabled := True;
end;

procedure TForm1.ButtonStep2Click(Sender: TObject);
var
  gmin, gmax: Integer;
begin
  (Sender as TButton).Enabled := False;

  ButtonStop.Enabled := True;
  gmin := StrToInt(LabelMinVal.Caption);
  gmax := StrToInt(LabelMaxVal.Caption);
  WorkThread1 := TWorkThread.Create(ListBoxFiles.Items, self, gmin, gmax);
end;

procedure TForm1.ButtonStopClick(Sender: TObject);
begin
  (Sender as TButton).Enabled := False;
  WorkThread1.Suicide;
  ButtonStep2.Enabled := True;
end;

initialization
  if not InitPngLib(PngLibname) then
  begin
    Application.MessageBox(@Format('Could not load %s', [PngLibname])[1], nil, 0);
    Halt(0);
  end;
finalization
  DeInitPngLib();
end.

